@@ -25,7 +25,7 @@
  * list of people on the gedit Team.  
  * See the ChangeLog files for a list of changes. 
  *
- * $Id: gedit-window-private.h 6650 2008-12-01 23:05:17Z pborelli $
+ * $Id: gedit-window-private.h 6755 2008-12-29 18:52:22Z jessevdk $
  */
 
 #ifndef __GEDIT_WINDOW_PRIVATE_H__
@@ -33,6 +33,7 @@
 
 #include "gedit/gedit-window.h"
 #include "gedit-prefs-manager.h"
+#include "gedit-message-bus.h"
 
 G_BEGIN_DECLS
 
@@ -46,7 +47,9 @@ struct _GeditWindowPrivate
 	GtkWidget      *bottom_panel;
 
 	GtkWidget      *hpaned;
-	GtkWidget      *vpaned;	
+	GtkWidget      *vpaned;
+	
+	GeditMessageBus *message_bus;	
 
 	/* statusbar and context ids for statusbar messages */
 	GtkWidget      *statusbar;	
