@@ -35,16 +35,18 @@
 #include "gedit-file-browser-error.h"
 #include "gedit-file-browser-utils.h"
 
-#define GEDIT_FILE_BROWSER_STORE_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE((object), GEDIT_TYPE_FILE_BROWSER_STORE, GeditFileBrowserStorePrivate))
+#define GEDIT_FILE_BROWSER_STORE_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE((object), \
+						     GEDIT_TYPE_FILE_BROWSER_STORE, \
+						     GeditFileBrowserStorePrivate))
 
-#define NODE_IS_DIR(node)				(FILE_IS_DIR((node)->flags))
-#define NODE_IS_HIDDEN(node)			(FILE_IS_HIDDEN((node)->flags))
-#define NODE_IS_TEXT(node)				(FILE_IS_TEXT((node)->flags))
-#define NODE_LOADED(node)				(FILE_LOADED((node)->flags))
-#define NODE_IS_FILTERED(node)			(FILE_IS_FILTERED((node)->flags))
-#define NODE_IS_DUMMY(node)				(FILE_IS_DUMMY((node)->flags))
+#define NODE_IS_DIR(node)		(FILE_IS_DIR((node)->flags))
+#define NODE_IS_HIDDEN(node)		(FILE_IS_HIDDEN((node)->flags))
+#define NODE_IS_TEXT(node)		(FILE_IS_TEXT((node)->flags))
+#define NODE_LOADED(node)		(FILE_LOADED((node)->flags))
+#define NODE_IS_FILTERED(node)		(FILE_IS_FILTERED((node)->flags))
+#define NODE_IS_DUMMY(node)		(FILE_IS_DUMMY((node)->flags))
 
-#define FILE_BROWSER_NODE_DIR(node)		((FileBrowserNodeDir *)(node))
+#define FILE_BROWSER_NODE_DIR(node)	((FileBrowserNodeDir *)(node))
 
 #define DIRECTORY_LOAD_ITEMS_PER_CALLBACK 100
 #define STANDARD_ATTRIBUTE_TYPES G_FILE_ATTRIBUTE_STANDARD_TYPE "," \
@@ -76,6 +78,7 @@ struct _AsyncNode
 {
 	FileBrowserNodeDir *dir;
 	GCancellable *cancellable;
+	GSList *original_children;
 };
 
 struct _FileBrowserNode 
@@ -188,10 +191,12 @@ static void next_files_async 				    (GFileEnumerator * enumerator,
 
 GEDIT_PLUGIN_DEFINE_TYPE_WITH_CODE (GeditFileBrowserStore, gedit_file_browser_store,
 			G_TYPE_OBJECT,
-			G_IMPLEMENT_INTERFACE (GTK_TYPE_TREE_MODEL,
-					       gedit_file_browser_store_iface_init)
-			G_IMPLEMENT_INTERFACE (GTK_TYPE_TREE_DRAG_SOURCE,
-					       gedit_file_browser_store_drag_source_init))
+			GEDIT_PLUGIN_IMPLEMENT_INTERFACE (gedit_file_browser_store_tree_model,
+							  GTK_TYPE_TREE_MODEL,
+							  gedit_file_browser_store_iface_init)
+			GEDIT_PLUGIN_IMPLEMENT_INTERFACE (gedit_file_browser_store_drag_source,
+							  GTK_TYPE_TREE_DRAG_SOURCE,
+							  gedit_file_browser_store_drag_source_init))
 
 /* Properties */
 enum {
@@ -224,10 +229,10 @@ gedit_file_browser_store_finalize (GObject * object)
 	GeditFileBrowserStore *obj = GEDIT_FILE_BROWSER_STORE (object);
 	GSList *item;
 
-	// Free all the nodes
+	/* Free all the nodes */
 	file_browser_node_free (obj, obj->priv->root);
 
-	// Cancel any asynchronous operations
+	/* Cancel any asynchronous operations */
 	for (item = obj->priv->async_handles; item; item = item->next)
 	{
 		AsyncData *data = (AsyncData *) (item->data);
@@ -237,8 +242,7 @@ gedit_file_browser_store_finalize (GObject * object)
 	}
 
 	g_slist_free (obj->priv->async_handles);
-	G_OBJECT_CLASS (gedit_file_browser_store_parent_class)->
-	    finalize (object);
+	G_OBJECT_CLASS (gedit_file_browser_store_parent_class)->finalize (object);
 }
 
 static void
@@ -247,9 +251,9 @@ set_gvalue_from_node (GValue          *value,
 {
 	gchar * uri;
 
-	if (node == NULL || !node->file)
+	if (node == NULL || !node->file) {
 		g_value_set_string (value, NULL);
-	else {
+	} else {
 		uri = g_file_get_uri (node->file);
 		g_value_take_string (value, uri);
 	}
@@ -531,9 +535,7 @@ gedit_file_browser_store_get_iter (GtkTreeModel * tree_model,
 {
 	gint * indices, depth, i;
 	FileBrowserNode * node;
-	FileBrowserNode * child;
 	GeditFileBrowserStore * model;
-	GSList * item;
 	gint num;
 
 	g_assert (GEDIT_IS_FILE_BROWSER_STORE (tree_model));
@@ -545,6 +547,8 @@ gedit_file_browser_store_get_iter (GtkTreeModel * tree_model,
 	node = model->priv->virtual_root;
 
 	for (i = 0; i < depth; ++i) {
+		GSList * item;
+
 		if (node == NULL)
 			return FALSE;
 
@@ -554,6 +558,8 @@ gedit_file_browser_store_get_iter (GtkTreeModel * tree_model,
 			return FALSE;
 
 		for (item = FILE_BROWSER_NODE_DIR (node)->children; item; item = item->next) {
+			FileBrowserNode * child;
+
 			child = (FileBrowserNode *) (item->data);
 
 			if (model_node_inserted (model, child)) {
@@ -583,12 +589,14 @@ static GtkTreePath *
 gedit_file_browser_store_get_path_real (GeditFileBrowserStore * model,
 					FileBrowserNode * node)
 {
-	GtkTreePath *path = gtk_tree_path_new ();
-	FileBrowserNode *check;
-	GSList *item;
+	GtkTreePath *path;
 	gint num = 0;
-	
+
+	path = gtk_tree_path_new ();
+
 	while (node != model->priv->virtual_root) {
+		GSList *item;
+
 		if (node->parent == NULL) {
 			gtk_tree_path_free (path);
 			return NULL;
@@ -597,6 +605,8 @@ gedit_file_browser_store_get_path_real (GeditFileBrowserStore * model,
 		num = 0;
 
 		for (item = FILE_BROWSER_NODE_DIR (node->parent)->children; item; item = item->next) {
+			FileBrowserNode *check;
+
 			check = (FileBrowserNode *) (item->data);
 
 			if (model_node_visibility (model, check)) {
@@ -626,16 +636,12 @@ static GtkTreePath *
 gedit_file_browser_store_get_path (GtkTreeModel * tree_model,
 				   GtkTreeIter * iter)
 {
-	FileBrowserNode *node;
-
 	g_return_val_if_fail (GEDIT_IS_FILE_BROWSER_STORE (tree_model), NULL);
 	g_return_val_if_fail (iter != NULL, NULL);
 	g_return_val_if_fail (iter->user_data != NULL, NULL);
 
-	node = (FileBrowserNode *) (iter->user_data);
-
 	return gedit_file_browser_store_get_path_real (GEDIT_FILE_BROWSER_STORE (tree_model), 
-						       node);
+						       (FileBrowserNode *) (iter->user_data));
 }
 
 static void
@@ -1131,14 +1137,14 @@ row_inserted (GeditFileBrowserStore * model,
 	gtk_tree_model_row_inserted (GTK_TREE_MODEL(model), copy, iter);
 	gtk_tree_path_free (copy);
 	
-	if (ref) {
+	if (ref)
+	{
 		gtk_tree_path_free (*path);
 
 		/* To restore the path, we get the path from the reference. But, since
 		   we inserted a row, the path will be one index further than the
 		   actual path of our node. We therefore call gtk_tree_path_prev */
 		*path = gtk_tree_row_reference_get_path (ref);
-	
 		gtk_tree_path_prev (*path);
 	}
 
@@ -1160,14 +1166,14 @@ row_deleted (GeditFileBrowserStore * model,
 static void
 model_refilter_node (GeditFileBrowserStore * model,
 		     FileBrowserNode * node,
-		     GtkTreePath * path)
+		     GtkTreePath ** path)
 {
 	gboolean old_visible;
 	gboolean new_visible;
 	FileBrowserNodeDir *dir;
 	GSList *item;
 	GtkTreeIter iter;
-	gboolean free_path = FALSE;
+	GtkTreePath *tmppath = NULL;
 	gboolean in_tree;
 
 	if (node == NULL)
@@ -1178,19 +1184,20 @@ model_refilter_node (GeditFileBrowserStore * model,
 
 	in_tree = node_in_tree (model, node);
 
-	if (path == NULL) {
+	if (path == NULL)
+	{
 		if (in_tree)
-			path = gedit_file_browser_store_get_path_real (model,
+			tmppath = gedit_file_browser_store_get_path_real (model,
 								       node);
 		else
-			path = gtk_tree_path_new_first ();
+			tmppath = gtk_tree_path_new_first ();
 
-		free_path = TRUE;
+		path = &tmppath;
 	}
 
 	if (NODE_IS_DIR (node)) {
 		if (in_tree)
-			gtk_tree_path_down (path);
+			gtk_tree_path_down (*path);
 
 		dir = FILE_BROWSER_NODE_DIR (node);
 
@@ -1201,7 +1208,7 @@ model_refilter_node (GeditFileBrowserStore * model,
 		}
 
 		if (in_tree)
-			gtk_tree_path_up (path);
+			gtk_tree_path_up (*path);
 	}
 
 	if (in_tree) {
@@ -1210,21 +1217,21 @@ model_refilter_node (GeditFileBrowserStore * model,
 		if (old_visible != new_visible) {
 			if (old_visible) {
 				node->inserted = FALSE;
-				row_deleted (model, path);
+				row_deleted (model, *path);
 			} else {
 				iter.user_data = node;
-				row_inserted (model, &path, &iter);
-				gtk_tree_path_next (path);
+				row_inserted (model, path, &iter);
+				gtk_tree_path_next (*path);
 			}
 		} else if (old_visible) {
-			gtk_tree_path_next (path);
+			gtk_tree_path_next (*path);
 		}
 	}
 	
 	model_check_dummy (model, node);
 
-	if (free_path)
-		gtk_tree_path_free (path);
+	if (tmppath)
+		gtk_tree_path_free (tmppath);
 }
 
 static void
@@ -1369,7 +1376,8 @@ file_browser_node_free (GeditFileBrowserStore * model,
  **/
 static void
 model_remove_node_children (GeditFileBrowserStore * model,
-			    FileBrowserNode * node, GtkTreePath * path,
+			    FileBrowserNode * node,
+			    GtkTreePath * path,
 			    gboolean free_nodes)
 {
 	FileBrowserNodeDir *dir;
@@ -1404,9 +1412,10 @@ model_remove_node_children (GeditFileBrowserStore * model,
 
 	list = g_slist_copy (dir->children);
 
-	for (item = list; item; item = item->next)
+	for (item = list; item; item = item->next) {
 		model_remove_node (model, (FileBrowserNode *) (item->data),
 				   path_child, free_nodes);
+	}
 
 	g_slist_free (list);
 	gtk_tree_path_free (path_child);
@@ -1425,15 +1434,14 @@ model_remove_node_children (GeditFileBrowserStore * model,
  * a node.
  **/
 static void
-model_remove_node (GeditFileBrowserStore * model, FileBrowserNode * node,
-		   GtkTreePath * path, gboolean free_nodes)
+model_remove_node (GeditFileBrowserStore * model,
+		   FileBrowserNode * node,
+		   GtkTreePath * path,
+		   gboolean free_nodes)
 {
 	gboolean free_path = FALSE;
 	FileBrowserNode *parent;
 
-	if (node == NULL)
-		return;
-
 	if (path == NULL) {
 		path =
 		    gedit_file_browser_store_get_path_real (model, node);
@@ -1515,13 +1523,14 @@ model_clear (GeditFileBrowserStore * model, gboolean free_nodes)
 	}
 }
 
-/* ====================================================================== */
-
 static void
 file_browser_node_unload (GeditFileBrowserStore * model,
 			  FileBrowserNode * node, gboolean remove_children)
 {
 	FileBrowserNodeDir *dir;
+	
+	if (node == NULL)
+		return;
 
 	if (!NODE_IS_DIR (node) || !NODE_LOADED (node))
 		return;
@@ -1565,10 +1574,15 @@ model_recomposite_icon_real (GeditFileBrowserStore * tree_model,
 
 	theme = gtk_icon_theme_get_default ();
 	
-	if (info)
-		icon = gedit_file_browser_utils_pixbuf_from_icon (g_file_info_get_icon (info), GTK_ICON_SIZE_MENU);
-	else
+	if (info) {
+		GIcon *gicon = g_file_info_get_icon (info);
+		if (gicon != NULL)
+			icon = gedit_file_browser_utils_pixbuf_from_icon (gicon, GTK_ICON_SIZE_MENU);
+		else
+			icon = NULL;
+	} else {
 		icon = gedit_file_browser_utils_pixbuf_from_file (node->file, GTK_ICON_SIZE_MENU);
+	}
 
 	if (node->icon)
 		g_object_unref (node->icon);
@@ -1649,14 +1663,14 @@ model_add_dummy_node (GeditFileBrowserStore * model,
 static void
 model_check_dummy (GeditFileBrowserStore * model, FileBrowserNode * node)
 {
-	FileBrowserNode *dummy;
-	FileBrowserNodeDir *dir;
-	GtkTreeIter iter;
-	GtkTreePath *path;
-	guint flags;
-
 	// Hide the dummy child if needed
 	if (NODE_IS_DIR (node)) {
+		FileBrowserNode *dummy;
+		GtkTreeIter iter;
+		GtkTreePath *path;
+		guint flags;
+		FileBrowserNodeDir *dir;
+
 		dir = FILE_BROWSER_NODE_DIR (node);
 
 		if (dir->children == NULL) {
@@ -1668,8 +1682,7 @@ model_check_dummy (GeditFileBrowserStore * model, FileBrowserNode * node)
 
 		if (!NODE_IS_DUMMY (dummy)) {
 			dummy = model_create_dummy_node (model, node);
-			dir->children =
-			    g_slist_prepend (dir->children, dummy);
+			dir->children = g_slist_prepend (dir->children, dummy);
 		}
 
 		if (!model_node_visibility (model, node)) {
@@ -1677,7 +1690,9 @@ model_check_dummy (GeditFileBrowserStore * model, FileBrowserNode * node)
 			    GEDIT_FILE_BROWSER_STORE_FLAG_IS_HIDDEN;
 			return;
 		}
-		// Temporarily set the node to invisible to check for real childs
+
+		/* Temporarily set the node to invisible to check
+		 * for real children */
 		flags = dummy->flags;
 		dummy->flags |= GEDIT_FILE_BROWSER_STORE_FLAG_IS_HIDDEN;
 
@@ -1717,7 +1732,8 @@ model_check_dummy (GeditFileBrowserStore * model, FileBrowserNode * node)
 }
 
 static void
-insert_node_sorted (GeditFileBrowserStore * model, FileBrowserNode * child,
+insert_node_sorted (GeditFileBrowserStore * model,
+		    FileBrowserNode * child,
 		    FileBrowserNode * parent)
 {
 	FileBrowserNodeDir *dir;
@@ -1738,19 +1754,18 @@ static void
 model_add_node (GeditFileBrowserStore * model, FileBrowserNode * child,
 		FileBrowserNode * parent)
 {
-	GtkTreeIter iter;
-	GtkTreePath *path;
-
-	// Add child to parents children
+	/* Add child to parents children */
 	insert_node_sorted (model, child, parent);
 
 	if (model_node_visibility (model, parent) &&
 	    model_node_visibility (model, child)) {
-		iter.user_data = child;
-		path =
-		    gedit_file_browser_store_get_path_real (model, child);
+		GtkTreeIter iter;
+		GtkTreePath *path;
 
-		// Emit row inserted
+		iter.user_data = child;
+		path = gedit_file_browser_store_get_path_real (model, child);
+
+		/* Emit row inserted */
 		row_inserted (model, &path, &iter);
 		gtk_tree_path_free (path);
 	}
@@ -1759,6 +1774,92 @@ model_add_node (GeditFileBrowserStore * model, FileBrowserNode * child,
 	model_check_dummy (model, child);
 }
 
+static void
+model_add_nodes_batch (GeditFileBrowserStore * model,
+		       GSList * children,
+		       FileBrowserNode * parent)
+{
+	GSList *sorted_children;
+	GSList *child;
+	GSList *prev;
+	GSList *l;
+	FileBrowserNodeDir *dir;
+
+	dir = FILE_BROWSER_NODE_DIR (parent);
+
+	sorted_children = g_slist_sort (children, (GCompareFunc) model->priv->sort_func);
+
+	child = sorted_children;
+	l = dir->children;
+	prev = NULL;
+
+	model_check_dummy (model, parent);
+
+	while (child) {
+		FileBrowserNode *node = child->data;
+		GtkTreeIter iter;
+		GtkTreePath *path;
+
+		/* reached the end of the first list, just append the second */
+		if (l == NULL) {
+
+			dir->children = g_slist_concat (dir->children, child);
+
+			for (l = child; l; l = l->next) {
+				if (model_node_visibility (model, parent) &&
+				    model_node_visibility (model, l->data)) {
+					iter.user_data = l->data;
+					path = gedit_file_browser_store_get_path_real (model, l->data);
+
+					// Emit row inserted
+					row_inserted (model, &path, &iter);
+					gtk_tree_path_free (path);
+				}
+
+				model_check_dummy (model, l->data);
+			}
+
+			break;
+		}
+
+		if (model->priv->sort_func (l->data, node) > 0) {
+			GSList *next_child;
+
+			if (prev == NULL) {
+				/* prepend to the list */
+				dir->children = g_slist_prepend (dir->children, child);
+			} else {
+				prev->next = child;
+			}
+
+			next_child = child->next;
+			prev = child;
+			child->next = l;
+			child = next_child;
+
+			if (model_node_visibility (model, parent) &&
+			    model_node_visibility (model, node)) {
+				iter.user_data = node;
+				path = gedit_file_browser_store_get_path_real (model, node);
+
+				// Emit row inserted
+				row_inserted (model, &path, &iter);
+				gtk_tree_path_free (path);
+			}
+
+			model_check_dummy (model, node);
+
+			/* try again at the same l position with the
+			 * next child */
+		} else {
+
+			/* Move to the next item in the list */
+			prev = l;
+			l = l->next;
+		}
+	}
+}
+
 static gchar const *
 backup_content_type (GFileInfo * info)
 {
@@ -1838,7 +1939,7 @@ file_browser_node_set_from_info (GeditFileBrowserStore * model,
 
 	if (isadded) {
 		path = gedit_file_browser_store_get_path_real (model, node);
-		model_refilter_node (model, node, path);
+		model_refilter_node (model, node, &path);
 		gtk_tree_path_free (path);
 		
 		model_check_dummy (model, node->parent);
@@ -1848,17 +1949,13 @@ file_browser_node_set_from_info (GeditFileBrowserStore * model,
 }
 
 static FileBrowserNode *
-model_file_exists (GeditFileBrowserStore * model, FileBrowserNode * parent,
-		  GFile * file)
+node_list_contains_file (GSList *children, GFile * file)
 {
 	GSList *item;
-	FileBrowserNode *node;
 
-	if (!NODE_IS_DIR (parent))
-		return NULL;
+	for (item = children; item; item = item->next) {
+		FileBrowserNode *node;
 
-	for (item = FILE_BROWSER_NODE_DIR (parent)->children; item;
-	     item = item->next) {
 		node = (FileBrowserNode *) (item->data);
 
 		if (node->file != NULL
@@ -1879,9 +1976,8 @@ model_add_node_from_file (GeditFileBrowserStore * model,
 	gboolean free_info = FALSE;
 	GError * error = NULL;
 
-	// Check if it already exists
-	if ((node = model_file_exists (model, parent, file)) == NULL) {	
-		if (!info) {
+	if ((node = node_list_contains_file (FILE_BROWSER_NODE_DIR (parent)->children, file)) == NULL) {
+		if (info == NULL) {
 			info = g_file_query_info (file,
 						  STANDARD_ATTRIBUTE_TYPES,
 						  G_FILE_QUERY_INFO_NONE,
@@ -1889,11 +1985,11 @@ model_add_node_from_file (GeditFileBrowserStore * model,
 						  &error);
 			free_info = TRUE;
 		}
-		
+	
 		if (!info) {
 			g_warning ("Error querying file info: %s", error->message);
 			g_error_free (error);
-			
+		
 			/* FIXME: What to do now then... */
 			node = file_browser_node_new (file, parent);
 		} else if (g_file_info_get_file_type (info) == G_FILE_TYPE_DIRECTORY) {
@@ -1904,7 +2000,7 @@ model_add_node_from_file (GeditFileBrowserStore * model,
 
 		file_browser_node_set_from_info (model, node, info, FALSE);
 		model_add_node (model, node, parent);
-		
+	
 		if (info && free_info)
 			g_object_unref (info);
 	}
@@ -1912,6 +2008,67 @@ model_add_node_from_file (GeditFileBrowserStore * model,
 	return node;
 }
 
+/* We pass in a copy of the list of parent->children so that we do
+ * not have to check if a file already exists among the ones we just
+ * added */
+static void
+model_add_nodes_from_files (GeditFileBrowserStore * model,
+			    FileBrowserNode * parent,
+			    GSList * original_children,
+			    GList * files)
+{
+	GList *item;
+	GSList *nodes = NULL;
+
+	for (item = files; item; item = item->next) {
+		GFileInfo *info = G_FILE_INFO (item->data);
+		GFileType type;
+		gchar const * name;
+		GFile * file;
+		FileBrowserNode *node;
+
+		type = g_file_info_get_file_type (info);
+
+		/* Skip all non regular, non directory files */
+		if (type != G_FILE_TYPE_REGULAR &&
+		    type != G_FILE_TYPE_DIRECTORY &&
+		    type != G_FILE_TYPE_SYMBOLIC_LINK) {
+			g_object_unref (info);
+			continue;
+		}		
+
+		name = g_file_info_get_name (info);
+
+		/* Skip '.' and '..' directories */
+		if (type == G_FILE_TYPE_DIRECTORY &&
+		    (strcmp (name, ".") == 0 ||
+		     strcmp (name, "..") == 0)) {
+			continue;
+		}
+
+		file = g_file_get_child (parent->file, name);
+
+		if ((node = node_list_contains_file (original_children, file)) == NULL) {
+
+			if (g_file_info_get_file_type (info) == G_FILE_TYPE_DIRECTORY) {
+				node = file_browser_node_dir_new (model, file, parent);
+			} else {
+				node = file_browser_node_new (file, parent);
+			}
+
+			file_browser_node_set_from_info (model, node, info, FALSE);
+
+			nodes = g_slist_prepend (nodes, node);
+		}
+
+		g_object_unref (file);
+		g_object_unref (info);
+	}
+
+	if (nodes)
+		model_add_nodes_batch (model, nodes, parent);
+}
+
 static FileBrowserNode *
 model_add_node_from_dir (GeditFileBrowserStore * model,
 			 FileBrowserNode * parent,
@@ -1920,7 +2077,7 @@ model_add_node_from_dir (GeditFileBrowserStore * model,
 	FileBrowserNode *node;
 
 	/* Check if it already exists */
-	if ((node = model_file_exists (model, parent, file)) == NULL) {	
+	if ((node = node_list_contains_file (FILE_BROWSER_NODE_DIR (parent)->children, file)) == NULL) {	
 		node = file_browser_node_dir_new (model, file, parent);
 		file_browser_node_set_from_info (model, node, NULL, FALSE);
 
@@ -1950,7 +2107,6 @@ parse_dot_hidden_file (FileBrowserNode *directory)
 	int i;
 	FileBrowserNodeDir * dir = FILE_BROWSER_NODE_DIR (directory);
 
-
 	/* FIXME: We only support .hidden on file: uri's for the moment.
 	 * Need to figure out if we should do this async or sync to extend
 	 * it to all types of uris.
@@ -2022,10 +2178,9 @@ on_directory_monitor_event (GFileMonitor * monitor,
 
 	switch (event_type) {
 	case G_FILE_MONITOR_EVENT_DELETED:
-		node = model_file_exists (dir->model, parent, file);
+		node = node_list_contains_file (dir->children, file);
 
 		if (node != NULL) {
-			// Remove the node
 			model_remove_node (dir->model, node, NULL, TRUE);
 		}
 		break;
@@ -2044,6 +2199,7 @@ static void
 async_node_free (AsyncNode *async)
 {
 	g_object_unref (async->cancellable);
+	g_slist_free (async->original_children);
 	g_free (async);
 }
 
@@ -2053,12 +2209,7 @@ model_iterate_next_files_cb (GFileEnumerator * enumerator,
 			     AsyncNode * async)
 {
 	GList * files;
-	GList * item;
 	GError * error = NULL;
-	GFileInfo * info;
-	GFileType type;
-	gchar const * name;
-	GFile * file;
 	FileBrowserNodeDir * dir = async->dir;
 	FileBrowserNode * parent = (FileBrowserNode *)dir;
 	
@@ -2074,16 +2225,25 @@ model_iterate_next_files_cb (GFileEnumerator * enumerator,
 			g_object_unref (dir->cancellable);
 			dir->cancellable = NULL;
 			
+/*
+ * FIXME: This is temporarly, it is a bug in gio:
+ * http://bugzilla.gnome.org/show_bug.cgi?id=565924
+ */
+#ifndef G_OS_WIN32
 			if (g_file_is_native (parent->file) && dir->monitor == NULL) {
 				dir->monitor = g_file_monitor_directory (parent->file, 
 									 G_FILE_MONITOR_NONE,
 									 NULL,
 									 NULL);
-				g_signal_connect (dir->monitor,
-						  "changed",
-						  G_CALLBACK (on_directory_monitor_event),
-						  parent);
+				if (dir->monitor != NULL)
+				{
+					g_signal_connect (dir->monitor,
+							  "changed",
+							  G_CALLBACK (on_directory_monitor_event),
+							  parent);
+				}
 			}
+#endif
 
 			model_check_dummy (dir->model, parent);
 			model_end_loading (dir->model, parent);
@@ -2107,30 +2267,7 @@ model_iterate_next_files_cb (GFileEnumerator * enumerator,
 		g_file_enumerator_close (enumerator, NULL, NULL);
 		async_node_free (async);
 	} else {
-		for (item = files; item; item = item->next) {
-			info = G_FILE_INFO (item->data);
-			type = g_file_info_get_file_type (info);
-			
-			/* Skip all non regular, non directory files */
-			if (type != G_FILE_TYPE_REGULAR &&
-			    type != G_FILE_TYPE_DIRECTORY &&
-			    type != G_FILE_TYPE_SYMBOLIC_LINK)
-				continue;
-
-			name = g_file_info_get_name (info);
-
-			/* Skip '.' and '..' directories */
-			if (type == G_FILE_TYPE_DIRECTORY &&
-			    (strcmp (name, ".") == 0 ||
-			     strcmp (name, "..") == 0))
-				continue;
-
-			file = g_file_get_child (parent->file, name);
-			model_add_node_from_file (dir->model, parent, file, info);
-
-			g_object_unref (file);
-			g_object_unref (info);
-		}
+		model_add_nodes_from_files (dir->model, parent, async->original_children, files);
 		
 		g_list_free (files);
 		next_files_async (enumerator, async);
@@ -2156,16 +2293,18 @@ model_iterate_children_cb (GFile * file,
 {
 	GError * error = NULL;
 	GFileEnumerator * enumerator;
+
+	if (g_cancellable_is_cancelled (async->cancellable))
+	{
+		async_node_free (async);
+		return;
+	}
 	
 	enumerator = g_file_enumerate_children_finish (file, result, &error);
-	
+
 	if (enumerator == NULL) {
-		/* Simply return if we were cancelled */
+		/* Simply return if we were cancelled or if the dir is not there */
 		FileBrowserNodeDir *dir = async->dir;
-		g_free (async);
-		
-		if (error->domain == G_IO_ERROR && error->code == G_IO_ERROR_CANCELLED)
-			return;
 		
 		/* Otherwise handle the error appropriately */
 		g_signal_emit (dir->model,
@@ -2176,10 +2315,7 @@ model_iterate_children_cb (GFile * file,
 
 		file_browser_node_unload (dir->model, (FileBrowserNode *)dir, TRUE);
 		g_error_free (error);
-	} else if (g_cancellable_is_cancelled (async->cancellable)) {
-		/* Check cancelled state manually */
 		async_node_free (async);
-		return;
 	} else {
 		next_files_async (enumerator, async);
 	}
@@ -2211,6 +2347,7 @@ model_load_directory (GeditFileBrowserStore * model,
 	AsyncNode *async = g_new (AsyncNode, 1);
 	async->dir = dir;
 	async->cancellable = g_object_ref (dir->cancellable);
+	async->original_children = g_slist_copy (dir->children);
 
 	/* Start loading async */
 	g_file_enumerate_children_async (node->file,
@@ -2243,7 +2380,7 @@ get_parent_files (GeditFileBrowserStore * model, GFile * file)
 
 static void
 model_fill (GeditFileBrowserStore * model, FileBrowserNode * node,
-	    GtkTreePath * path)
+	    GtkTreePath ** path)
 {
 	gboolean free_path = FALSE;
 	GtkTreeIter iter = {0,};
@@ -2252,19 +2389,19 @@ model_fill (GeditFileBrowserStore * model, FileBrowserNode * node,
 
 	if (node == NULL) {
 		node = model->priv->virtual_root;
-		path = gtk_tree_path_new ();
+		*path = gtk_tree_path_new ();
 		free_path = TRUE;
 	}
 
-	if (path == NULL) {
-		path =
+	if (*path == NULL) {
+		*path =
 		    gedit_file_browser_store_get_path_real (model, node);
 		free_path = TRUE;
 	}
 
 	if (!model_node_visibility (model, node)) {
 		if (free_path)
-			gtk_tree_path_free (path);
+			gtk_tree_path_free (*path);
 
 		return;
 	}
@@ -2273,12 +2410,12 @@ model_fill (GeditFileBrowserStore * model, FileBrowserNode * node,
 		/* Insert node */
 		iter.user_data = node;
 		
-		row_inserted(model, &path, &iter);
+		row_inserted(model, path, &iter);
 	}
 
 	if (NODE_IS_DIR (node)) {
 		/* Go to the first child */
-		gtk_tree_path_down (path);
+		gtk_tree_path_down (*path);
 
 		for (item = FILE_BROWSER_NODE_DIR (node)->children; item;
 		     item = item->next) {
@@ -2288,18 +2425,18 @@ model_fill (GeditFileBrowserStore * model, FileBrowserNode * node,
 				model_fill (model, child, path);
 
 				/* Increase path for next child */
-				gtk_tree_path_next (path);
+				gtk_tree_path_next (*path);
 			}
 		}
 
 		/* Move back up to node path */
-		gtk_tree_path_up (path);
+		gtk_tree_path_up (*path);
 	}
 	
 	model_check_dummy (model, node);
 
 	if (free_path)
-		gtk_tree_path_free (path);
+		gtk_tree_path_free (*path);
 }
 
 static void
@@ -2312,6 +2449,7 @@ set_virtual_root_from_node (GeditFileBrowserStore * model,
 	FileBrowserNodeDir *dir;
 	GSList *item;
 	GSList *copy;
+	GtkTreePath *empty = NULL;
 
 	prev = node;
 	next = prev->parent;
@@ -2376,7 +2514,7 @@ set_virtual_root_from_node (GeditFileBrowserStore * model,
 
 	/* Now finally, set the virtual root, and load it up! */
 	model->priv->virtual_root = node;
-	model_fill (model, NULL, NULL);
+	model_fill (model, NULL, &empty);
 
 	if (!NODE_LOADED (node))
 		model_load_directory (model, node);
@@ -2460,10 +2598,9 @@ gedit_file_browser_store_error_quark (void)
 {
 	static GQuark quark = 0;
 
-	if (G_UNLIKELY (quark == 0))
-		quark =
-		    g_quark_from_static_string
-		    ("gedit_file_browser_store_error");
+	if (G_UNLIKELY (quark == 0)) {
+		quark = g_quark_from_string ("gedit_file_browser_store_error");
+	}
 
 	return quark;
 }
@@ -2704,12 +2841,12 @@ gedit_file_browser_store_set_virtual_root (GeditFileBrowserStore * model,
 GeditFileBrowserStoreResult
 gedit_file_browser_store_set_virtual_root_from_string
     (GeditFileBrowserStore * model, gchar const *root) {
-	GFile * file = g_file_new_for_uri (root);
-	gchar *str, *str1;
+	GFile *file;
 
 	g_return_val_if_fail (GEDIT_IS_FILE_BROWSER_STORE (model),
 			      GEDIT_FILE_BROWSER_STORE_RESULT_NO_CHANGE);
 
+	file = g_file_new_for_uri (root);
 	if (file == NULL) {
 		g_warning ("Invalid uri (%s)", root);
 		return GEDIT_FILE_BROWSER_STORE_RESULT_NO_CHANGE;
@@ -2733,8 +2870,10 @@ gedit_file_browser_store_set_virtual_root_from_string
 	}
 
 	if (!g_file_has_prefix (file, model->priv->root->file)) {
-		str = gedit_file_browser_utils_file_display (model->priv->root->file);
-		str1 = gedit_file_browser_utils_file_display (file);
+		gchar *str, *str1;
+
+		str = g_file_get_parse_name (model->priv->root->file);
+		str1 = g_file_get_parse_name (file);
 
 		g_warning
 		    ("Virtual root (%s) is not below actual root (%s)",
@@ -3024,7 +3163,7 @@ gedit_file_browser_store_refresh (GeditFileBrowserStore * model)
 {
 	g_return_if_fail (GEDIT_IS_FILE_BROWSER_STORE (model));
 
-	if (model->priv->root == NULL)
+	if (model->priv->root == NULL || model->priv->virtual_root == NULL)
 		return;
 
 	/* Clear the model */
@@ -3186,8 +3325,8 @@ file_deleted (IdleDelete * data)
 {
 	FileBrowserNode * node;
 	node = model_find_node (data->model, NULL, data->file);
-	    	
-	if (node)
+
+	if (node != NULL)
 		model_remove_node (data->model, node, NULL, TRUE);
 	
 	return FALSE;
